/*
    Copyright (c) 2002 Florian Wesch <fw@dividuum.de>. All Rights Reserved.     
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
*/    


#include <stdio.h>
#include "common.h"
#include "bfasm.h"

/* 
   TODO

   * Flexibleres Speichermanagement fr temporre Variablen
*/   

// Funktionen

BFAssemblerContext *BF_AsmInit(BFADDR freefortemp) {
    BFAssemblerContext *ctx;
    ctx = (BFAssemblerContext*)malloc(sizeof(BFAssemblerContext));

    if (ctx == NULL)
        fatalerror("Huch. keinen Speicher fr den Assembler?");

    ctx->looplevel = 0;
    ctx->first = NULL;
    ctx->cur = NULL;
    ctx->freefortemp = freefortemp;

    return ctx;
}

void BF_AllocNewInst(BFAssemblerContext *ctx) {
    BFInstruction* new = (BFInstruction*)malloc(sizeof(BFInstruction));
    if (new == NULL)
        fatalerror("Huch. keinen Speicher fr neue Anweisung?");

    if (ctx->first == NULL) {
        ctx->first = new;
        ctx->cur = ctx->first;
    } else {
        ctx->cur->next = new;
        ctx->cur = ctx->cur->next;
    }
    ctx->cur->next = NULL;
}

void BF_AssemblerDump(BFAssemblerContext *ctx) {
    BFInstruction *inst = ctx->first;
    while (inst != NULL) {
        switch (inst->type) {
            case BFInst:
                printf("%c ", inst->val.bfinst);
                break;
            case BFMove:
                printf("TO(%d) ", inst->val.moveto);
                break;
        }
        inst = inst->next;
    }
}

#define MoveFromTo(source, dest) {      \
    int i;                              \
    if (dest < source) {                \
        for (i=source; i>dest; i--)     \
            fprintf(out, "<");          \
    } else if (dest > source) {         \
        for (i=source; i<dest; i++)     \
            fprintf(out, ">");          \
    }                                   \
}

void BF_Assemble(BFAssemblerContext *ctx, FILE *out) {
    BFADDR lastpos=0; // IP steht am Anfang auf 0
    BFInstruction *inst = ctx->first;

    fprintf(out, "[ generated by bfcc "BFCCVERSION" ]\n");
    while (inst != NULL) {
        switch (inst->type) {
            case BFInst:
                fprintf(out, "%c", inst->val.bfinst);
                break;
            case BFMove:
                MoveFromTo(lastpos, inst->val.moveto);
                lastpos = inst->val.moveto;
                break;
        }
        inst = inst->next;
    }
    MoveFromTo(lastpos, 0); // Am Ende wieder auf 0 zurck
}

BFADDR BF_Newvar(BFAssemblerContext *ctx) {
    BFADDR nextfree = ctx->freefortemp;
    ctx->freefortemp++;
    
    if (ctx->freefortemp >= MAXMEMSIZE)
        fatalerror("freefortemp > %d\n", MAXMEMSIZE);
    
    //info("Neue Variable @%d\n", nextfree);

    return nextfree;
}

void BF_Removevar(BFAssemblerContext *ctx, BFADDR mem_to_free) {
    if (mem_to_free == ctx->freefortemp-1)
        ctx->freefortemp--;
    else
        fatalerror("Whoops: ctx->freefortemp-1 == %d, mem_to_free == %d", 
                    ctx->freefortemp-1,
                    mem_to_free);
    
    //info("Tte Variable @%d\n", mem_to_free);
}

void BF_StartLoop(BFAssemblerContext *ctx) {
    ctx->looplevel++;
    
    if (ctx->looplevel >= MAXLOOPLEVEL)
        fatalerror("Maxlooplevel berschritten");
    
}

void BF_EndLoop(BFAssemblerContext *ctx) {
    ctx->looplevel--;

    if (ctx->looplevel < 0)
        fatalerror("Looplevel < 0: Interer Fehler!?!");
}

/*    
    to(dest)
        < oder >
*/    
BFINST1(to, BFADDR dest) {    
    BFMOVE(dest);
}

/*
    zero(dest)
        to(dest) [ - ] 
*/        
BFINST1(zero, BFADDR dest) { 
    while_begin(dest);
        BFINST(BF_OPDECV); 
    while_end(dest);
}

/*
    print(mem)
        to(mem) . 
*/        
BFINST1(print, BFADDR mem) { 
    BFCALL1(to, mem);
    BFINST(BF_OPOUT); 
}

/*
    input(dest)
        to(dest) ,
*/        
BFINST1(input, BFADDR dest) { 
    BFCALL1(to, dest);
    BFINST(BF_OPIN); 
}

/*
    add_const(dest, {constvalue})
        to(dest) + {*constvalue}
*/        
BFINST2(add_const, BFADDR dest, BFVALUE constvalue) {
    int i;
    BFCALL1(to, dest);
    for (i=0; i<constvalue; i++)
        BFINST(BF_OPINCV);
}

/*
    set_const(dest, {constvalue})
        zero(dest) add_const(dest, {constvalue})
*/
BFINST2(set_const, BFADDR dest, BFVALUE constvalue) {
    BFCALL1(zero, dest);
    BFCALL2(add_const, dest, constvalue);
}

/*
    inc(dest)
        to(dest) + 
*/    
BFINST1(inc, BFADDR dest) {
    BFCALL1(to, dest);
    BFINST(BF_OPINCV);
}

/*
    dec(dest)
        to(dest) -
*/
BFINST1(dec, BFADDR dest) {
    BFCALL1(to, dest);
    BFINST(BF_OPDECV);
}

/*
    add_destroy(dest, value)
        to(value) [ inc(dest) dec(value) ]
*/
BFINST2(add_destroy, BFADDR dest, BFADDR value) {
    while_begin(value)
        BFCALL1(inc, dest);
        BFCALL1(dec, value);
    while_end(value);
}

/*
    substract_destroy(dest, value)
        to(value) [ dec(dest) dec(value) ]
*/
BFINST2(substract_destroy, BFADDR dest, BFADDR value) {
    while_begin(value)
        BFCALL1(dec, dest);
        BFCALL1(dec, value);
    while_end(value);
}


/*
    move_destroy(source, dest)
        zero(dest) add_destroy(dest, source)
*/
BFINST2(move_destroy, BFADDR source, BFADDR dest) {
    BFCALL1(zero, dest);
    BFCALL2(add_destroy, dest, source);
}

/*
    move2-destroy(source, dest1, dest2)
        zero(dest1) zero(dest2) to(source) [ inc(dest1) inc(dest2) dec(source) ]
*/        
BFINST3(move2_destroy, BFADDR source, BFADDR dest1, BFADDR dest2) {
    BFCALL1(zero, dest1);
    BFCALL1(zero, dest2);
    while_begin(source);
        BFCALL1(inc, dest1);
        BFCALL1(inc, dest2);
        BFCALL1(dec, source);
    while_end(source);
}

/*
    copy(source, dest)
        newvar(temp) move2-destroy(source, dest, temp) move-destroy(temp, source) removevar(temp)
*/
BFINST2(copy, BFADDR source, BFADDR dest) {
    BFADDR temp = BFNEWVAR;
        BFCALL3(move2_destroy, source, dest, temp);
        BFCALL2(move_destroy, temp, source);
    BFREMOVEVAR(temp);
}

/*
    mul(memory, value) 
        newvar(temp1) newvar(temp2) zero(temp1) to(memory) 
        [ copy(value, temp2) add-destroy(temp1, temp2) dec(memory) ] 
        move(temp1, memory) removevar(temp1) removevar(temp2)   
*/
BFINST2(mul, BFADDR memory, BFADDR value) {
    BFADDR temp1 = BFNEWVAR;
    BFADDR temp2 = BFNEWVAR;
        BFCALL1(zero, temp1);
        while_begin(memory);
            BFCALL2(copy, value, temp2);
            BFCALL2(add_destroy, temp1, temp2);
            BFCALL1(dec, memory);
        while_end(memory);
        BFCALL2(move_destroy, temp1, memory);
    BFREMOVEVAR(temp2);
    BFREMOVEVAR(temp1);
}

/*
    mulconst(memory, {constvalue})
        newvar(temp) setconst(temp, {constvalue}) mul(memory, temp) removevar(temp) 
*/        
BFINST2(mul_const, BFADDR memory, BFVALUE constvalue) {
    BFADDR temp = BFNEWVAR;
    BFCALL2(set_const, temp, constvalue);
    BFCALL2(mul, memory, temp);
    BFREMOVEVAR(temp);
}

/*
    not(mem)
        newvar(temp) setconst(temp, 1) if(mem, <zero(temp)>) move-destroy(temp, mem) removevar(temp)
*/
BFINST1(not, BFADDR mem) {
    BFADDR temp = BFNEWVAR;
    BFCALL2(set_const, temp, 1);
    if_begin(mem)
        BFCALL1(zero, temp);
    if_end;
    BFCALL2(move_destroy, temp, mem);
    BFREMOVEVAR(temp);
}

/*
    and(source1, source2, dest)
        zero(dest) if(source1, <copy(source2, dest)>)
*/
BFINST3(and, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL1(zero, dest);
    if_begin(source1);
        BFCALL2(copy, source2, dest);
    if_end;
}

/*
    or(source1, source2, dest)
        newvar(temp) zero(dest) copy(source1, temp) add-destroy(dest, temp) 
        copy(source2, temp) add-destroy(dest, temp) removevar(temp)
*/        
BFINST3(or, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFADDR temp = BFNEWVAR;
    BFCALL1(zero, dest);
    BFCALL2(copy, source1, temp);
    BFCALL2(add_destroy, dest, temp);
    BFCALL2(copy, source2, temp);
    BFCALL2(add_destroy, dest, temp);
    BFREMOVEVAR(temp);
}

/*
    substractToNull-destroy(source1, source2)
        newvar(temp) and(source1, source2, temp)
        to(temp) [ dec(source1) dec(source2) and(source1, source2, temp) to(temp) ] removevar(temp)
*/
BFINST2(substractToNull_destroy, BFADDR source1, BFADDR source2) {
    BFADDR temp = BFNEWVAR;
    BFCALL3(and, source1, source2, temp);
    while_begin(temp);
        BFCALL1(dec, source1);
        BFCALL1(dec, source2);
        BFCALL3(and, source1, source2, temp);
    while_end(temp);
    BFREMOVEVAR(temp);
}

/*
    NotEqual-destroy(source1, source2, dest) 
        substractToNull-destroy(source1, source2)
        or(source1, source2, dest)
*/
BFINST3(notEqual_destroy, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL2(substractToNull_destroy, source1, source2);
    BFCALL3(or, source1, source2, dest);
}

/*
    Equal-destroy(source1, source2, dest)
        NotEqual(source1, source2, dest) not(dest)
*/
BFINST3(Equal_destroy, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL3(notEqual_destroy, source1, source2, dest);
    BFCALL1(not, dest);
}

/*
    Greater-destroy(source1, source2, dest)
        substractToNull-destroy(source1, source2)
        move-destroy(source1, dest)   
*/
BFINST3(Greater_destroy, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL2(substractToNull_destroy, source1, source2);
    BFCALL2(move_destroy, source1, dest);
}

/*
    Less-destroy(source1, source2, dest)
        substractToNull-destroy(source1, source2)
        move-destroy(source2, dest)    
*/
BFINST3(Less_destroy, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL2(substractToNull_destroy, source1, source2);
    BFCALL2(move_destroy, source2, dest);
}

/*
    GreaterOrEqual-destroy(source1, source2, dest) 
        inc(source1) Greater-destroy(source1, source2, dest)
*/        
BFINST3(GreaterOrEqual_destroy, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL1(inc, source1);
    BFCALL3(Greater_destroy, source1, source2, dest);
}

/*
    LessOrEqual-destroy(source1, source2, dest) 
        inc(source2) Less-destroy(source1, source2, dest)
*/        
BFINST3(LessOrEqual_destroy, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL1(inc, source2);
    BFCALL3(Less_destroy, source1, source2, dest);
}

