/*
    Copyright (c) 2002 Florian Wesch <fw@dividuum.de>. All Rights Reserved.     
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
*/    


#include <stdio.h>
#include <unistd.h>
#include "compiler.h"
#include "common.h"

void yyparse();
FILE *yyin;

void yyerror(char *s) { 
    fatalerror("%s\n", s); 
} 

void usage() {
    fprintf(stderr, "\n");
    fprintf(stderr, "[[ dividuums dumblanguage -> brainfuck compiler "BFCCVERSION" ]]\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Usage: bfcc [options] file\n");
    fprintf(stderr, "Options: -h            Help\n");
    fprintf(stderr, "         -O <level>    Optimizing level, 0=none 1=some 2=TODO,  default 1\n");
    fprintf(stderr, "         -o <file>     Place the output into <file>\n");
    exit(42);
}

int main(int argc, char **argv) { 
    char *infile  = "-";
    char *outfile = "-";
    FILE *in    = stdin;
    FILE *out   = stdout;
    int  optimizing = 1;
    int  opt;

    while ((opt = getopt(argc, argv, "O:o:h"))!=-1) {
        switch (opt) {
            case 'h':
                usage();
                break;
            case 'o':
                outfile = optarg;
                break;
            case 'O':
                optimizing = atoi(optarg);
                break;
        }
    }

    if (optind == argc-1) { 
        infile = argv[argc-1];
    } 

    if (strcmp(infile, "-") != 0 && !(in=fopen(infile,"r")))
        fatalerror("kann input file %s nicht ffnen\n", infile);

    if (strcmp(outfile, "-") != 0 && !(out=fopen(outfile,"w")))
        fatalerror("kann output file %s nicht ffnen\n", outfile);
        
    yyin = in;

    info("Parsing...\n");
    
    st = STInit();
    yyparse(); 
    
    info("Assigning memory...\n");
    
    STAssignLocation(st);
    
    fclose(in);
    
    STDump(st);

    info("Optimizing...\n");

    //dumpdotty(rootNode);
    if (optimizing > 0) 
        optimize(rootNode);
    //dumpdotty(rootNode);

    info("Compiling...\n");
    
    ctx = BF_AsmInit(st->lastmempos);
    compile(rootNode, INVALIDADDR);
    freeNode(rootNode);

    info("Assembling...\n");
    
	BF_Assemble(ctx, out);
    return 0; 
} 



