/*
    Copyright (c) 2002 Florian Wesch <fw@dividuum.de>. All Rights Reserved.     
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
*/    


#include <stdlib.h>
#include <stdarg.h>
#include "node.h"
#include "common.h"

nodeType *newNode(nodeEnum type) {
    nodeType *p; 
    
    /* allocate node */ 
    if ((p = malloc(sizeof(nodeType))) == NULL) 
        fatalerror("out of memory"); 
    
    p->type = type;
    return p;
}

nodeType *con(int value) { 
    nodeType *p = newNode(typeCon);

    p->as.con.value = value; 
    
    return p; 
} 


nodeType *id(char *name, struct nodeTypeTag *index) {
    nodeType *p = newNode(typeId);
    Symbol *s;
    
    /* copy information */ 
    p->as.id.varname = name; 
    p->as.id.index   = index; 

    s = GetSymbol(st, name);
    if (index->type == typeCon) {
        if (!s) {
            NewSymbol(st, name, index->as.con.value+1);
        } else {
            SymbolGrowSize(st, name, index->as.con.value);
        }
    } else if (!s) {
        fatalerror("%s used with index expression, but storage size of %s unknown", name, name);
    }
    
    return p; 
} 

nodeType *opr(int oper, int nops, ...) { 
    nodeType *p = newNode(typeOpr);
    va_list ap; 
    int i; 
    
    /* copy information */ 
    p->as.opr.oper = oper; 
    p->as.opr.nops = nops; 
    
    va_start(ap, nops); 
    for (i = 0; i < nops; i++)
        p->as.opr.op[i] = va_arg(ap, nodeType*); 
    va_end(ap); 

    return p; 
} 

void freeNodeContent(nodeType *p) { 
    if (!p) 
        return; 

    if (p->type == typeId) {
        free(p->as.id.varname);
    }

} 
 
void freeNode(nodeType *p) { 
    int i; 
    
    if (!p) 
        return; 

    // Unternodes freeen
    if (p->type == typeId) {
        freeNode(p->as.id.index); 
    } else if (p->type == typeOpr) {
        for (i = 0; i < p->as.opr.nops; i++) {
            freeNode(p->as.opr.op[i]); 
        }
    }

    freeNodeContent(p);
    free(p); 
} 

void freeOneNode(nodeType *p) { 
    if (!p) 
        return; 

    freeNodeContent(p);
    free(p); 
} 

