/*
    Copyright (c) 2002 Florian Wesch <fw@dividuum.de>. All Rights Reserved.     
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
*/    


#ifndef NODE_H
#define NODE_H

typedef enum { typeCon, typeId, typeOpr } nodeEnum; 

/* constants */ 
typedef struct { 
    int value;                  /* value of constant */ 
} ConNode; 

/* identifiers */ 
typedef struct { 
    char *varname;              /* variable name */ 
    struct nodeTypeTag *index;   /* index (for arrays) */
} IdNode; 

/* operators */ 
typedef struct { 
    int oper;                   /* operator */ 
    int nops;                   /* number of operands */ 
    struct nodeTypeTag *op[3];  /* operands */ 
} OprNode; 


typedef struct nodeTypeTag { 
    nodeEnum    type;       /* type of node */ 
    union {
        ConNode     con;    /* constants */ 
        IdNode      id;     /* identifiers */ 
        OprNode     opr;    /* operators */ 
    } as;
} nodeType; 

nodeType *opr(int oper, int nops, ...); 
nodeType *id(char *name, struct nodeTypeTag *index);
nodeType *con(int value); 

void freeNodeContent(nodeType *p);
void freeOneNode(nodeType *p);
void freeNode(nodeType *p); 

// Ersetzt dest durch replacement. replacement wird 
// danach gelscht, ALLERDINGS nur replacement und 
// nicht darin enthaltene Nodes (index, op[])
#define ReplaceNode(dest, replacement) {    \
    nodeType *tmp = replacement;            \
    *(dest) = *(replacement);               \
    freeOneNode(tmp);                       \
}                                           

// Ersetzt gegebene Node durch eine Node ohne Funktion
// Vorsicht: Falls dest eine ID oder eine Opr waren, 
// so werden in der Node gespeicherte Nodes (index, op[])
// nicht gelscht
#define MakeDummyNode(dest) {               \
    freeNodeContent(dest);                  \
    dest->as.opr.oper = ';';                \
    dest->as.opr.nops = 0;                  \
}

// Makros fr einfachen zugriff auf nodes
#define N_TYPE(n)           ((n)->type)
#define N_VALUE(n)          ((n)->as.con.value)
#define N_VARNAME(n)        ((n)->as.id.varname)
#define N_OPER(n)           ((n)->as.opr.oper)
#define N_NOPS(n)           ((n)->as.opr.nops)
#define N_OP(n, num)        ((n)->as.opr.op[num])
#define N_OPTYPE(n, num)    (N_TYPE(N_OP(n, (num))))
#define N_OPVALUE(n, num)   (N_VALUE(N_OP(n, (num))))
#define N_OPVARNAME(n, num) (N_VARNAME(N_OP(n, (num))))

#endif // NODE_H
