/*
    Copyright (c) 2002 Florian Wesch <fw@dividuum.de>. All Rights Reserved.     
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
*/    


%{ 
#include <stdio.h> 
#include <stdlib.h> 
#include <stdarg.h> 
#include "node.h" 
#include "compiler.h" 
#include "common.h"    

%} 


%union { 
    int intValue;       /* integer value */ 
    char *varName;      /* symbol table index */ 
    nodeType *nPtr;     /* node pointer */ 
}; 

%token WHILE IF PRINT INPUT

%token <intValue> INTEGER 
%token <varName> VARIABLE 
%type <nPtr> stmt expr stmt_list 


%nonassoc IFX 
%nonassoc ELSE 
// %nonassoc UMINUS 

%left AND OR
%left GE LE EQ NE '>' '<' 
%left '+' '-' 
%left '*' '/' 

%% 

program: 
        stmt_list                           { rootNode=$1; }
        | /* NULL */
        ;

stmt_list: 
          stmt                              { $$ = $1; } 
        | stmt_list stmt                    { $$ = opr(';', 2, $1, $2); } 
        ; 

stmt: 
          ';'                               { $$ = opr(';', 2, NULL, NULL); } 
      //| expr ';'                          { $$ = $1; }  // Was geht durchs auskommentieren verloren?
      //| INPUT VARIABLE ';'                { $$ = opr(INPUT, 1, id($2, con(0))); }
      //| INPUT VARIABLE '[' expr ']' ';'   { $$ = opr(INPUT, 1, id($2, $4)); }
        | PRINT expr ';'                    { $$ = opr(PRINT, 1, $2); } 
        | VARIABLE '=' expr ';'             { $$ = opr('=', 2, id($1, con(0)), $3); } 
        | VARIABLE '[' expr ']' '=' expr ';'{ $$ = opr('=', 2, id($1, $3), $6); } 
        | WHILE '(' expr ')' stmt           { $$ = opr(WHILE, 2, $3, $5); } 
        | IF '(' expr ')' stmt %prec IFX    { $$ = opr(IF, 2, $3, $5); } 
        | IF '(' expr ')' stmt ELSE stmt    { $$ = opr(IF, 3, $3, $5, $7); } 
        | '{' stmt_list '}'                 { $$ = $2; } 
        ; 

expr: 
          INTEGER               { $$ = con($1); } 
        | INPUT                 { $$ = opr(INPUT, 0); }
        | VARIABLE              { $$ = id($1, con(0)); } 
        | VARIABLE '[' expr ']' { $$ = id($1, $3); } 
      //| '-' expr %prec UMINUS { $$ = opr(UMINUS, 1, $2); } 
        | expr '+' expr         { $$ = opr('+', 2, $1, $3); } 
        | expr '-' expr         { $$ = opr('-', 2, $1, $3); } 
        | expr '*' expr         { $$ = opr('*', 2, $1, $3); } 
        | expr '/' expr         { $$ = opr('/', 2, $1, $3); } 
        | expr '<' expr         { $$ = opr('<', 2, $1, $3); } 
        | expr '>' expr         { $$ = opr('>', 2, $1, $3); } 
        | expr GE expr          { $$ = opr(GE, 2, $1, $3); } 
        | expr LE expr          { $$ = opr(LE, 2, $1, $3); } 
        | expr NE expr          { $$ = opr(NE, 2, $1, $3); } 
        | expr EQ expr          { $$ = opr(EQ, 2, $1, $3); } 
        | expr OR expr          { $$ = opr(OR, 2, $1, $3); } 
        | expr AND expr         { $$ = opr(AND, 2, $1, $3); } 
        | '(' expr ')'          { $$ = $2; } 
        ; 

 %% 

