/*
    Copyright (c) 2002 Florian Wesch <fw@dividuum.de>. All Rights Reserved.     
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
*/    


%{ 
#include <stdlib.h> 
#include <string.h>
#include "node.h"
#include "y.tab.h" 
#include "common.h"
%} 

NEWLINE ("\r"|"\n"|"\r\n"|"\n\r")

%% 

"//"[^\r\n]*{NEWLINE} ; // Kommentare weg

"&&"            return AND; 
"||"            return OR; 
">="            return GE; 
"<="            return LE; 
"=>"            return GE; 
"=<"            return LE; 
"=="            return EQ; 
"!="            return NE; 
"<>"            return NE; 
"while"         return WHILE; 
"if"            return IF; 
"else"          return ELSE; 
"print"         return PRINT; 
"input"         return INPUT;
[ \t]           ;
{NEWLINE}       /* lineno++ */;

0[0-1]+b {
        char cur;
        int val=0;
        int mul=1;
        for (cur = strlen(yytext)-2; cur>=1; cur--) {
                val+=((*(yytext+cur)=='0')?0:1)*mul;
                mul<<=1;
        }
        yylval.intValue=val;
        return INTEGER;
}

0x[0-9a-fA-F]+  {
        char cur;
        char ziffer;
        int val=0;
        int mul=1;
        for (cur = strlen(yytext)-1; cur>=2; cur--) {
                ziffer=*(yytext+cur);
                //printf("akutelle ziffer: %c mul=%d\n", ziffer, mul);
                if (ziffer>='a' && ziffer<='f') 
                        val+=(10+ziffer-'a')*mul;
                else if (ziffer>='A' && ziffer<='F') 
                        val+=(10+ziffer-'A')*mul;
                else if (ziffer>='0' && ziffer<='9') 
                        val+=(ziffer-'0')*mul;
                mul<<=4;
        }
        //printf("hexzahl gefunden: %s = %d\n", yytext, val);
        yylval.intValue=val;
        return INTEGER;
}

'.' {
        yylval.intValue = yytext[1];
        return INTEGER;
}

[0-9]+ { 
        yylval.intValue = atoi(yytext); 
        return INTEGER; 
} 

[a-zA-Z][a-zA-Z0-9]* {  
        yylval.varName = strdup(yytext); 
        return VARIABLE; 
} 

[-()<>=+*/;{}.\[\]] { 
        return *yytext; 
} 

.       {
            printf("%c %02x\n", *yytext, *yytext);
            yyerror("Unknown character"); 
}

%% 

int yywrap(void) { 
    return 1; 
} 
 

